@@ -1,27 +1,28 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
+import time
+import math
+
 from xpra.log import Logger
 focuslog = Logger("focus")
 workspacelog = Logger("workspace")
 log = Logger("window")
-keylog = Logger("window", "keyboard")
+keylog = Logger("keyboard")
 
-from xpra.util import AdHocStruct, nn
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk
+from xpra.util import AdHocStruct, nn, bytestostr
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_cairo, import_pixbufloader
+from xpra.gtk_common.gtk_util import pixbuf_new_from_data, COLORSPACE_RGB
 from xpra.client.client_window_base import ClientWindowBase
-gtk = import_gtk()
-gdk = import_gdk()
-
-import os
-import cairo
-import time
-import math
-
+gtk     = import_gtk()
+gdk     = import_gdk()
+cairo   = import_cairo()
+PixbufLoader = import_pixbufloader()
 
 CAN_SET_WORKSPACE = False
 HAS_X11_BINDINGS = False
@@ -29,7 +30,7 @@ if os.name=="posix":
     try:
         from xpra.x11.gtk_x11.prop import prop_get, prop_set
         from xpra.x11.bindings.window_bindings import constants, X11WindowBindings  #@UnresolvedImport
-        from xpra.x11.gtk_x11.error import trap
+        from xpra.gtk_common.error import trap
         X11Window = X11WindowBindings()
         HAS_X11_BINDINGS = True
 
@@ -47,7 +48,6 @@ if os.name=="posix":
     except ImportError, e:
         pass
 
-
 #optional module providing faster handling of premultiplied argb:
 try:
     from xpra.codecs.argb.argb import unpremultiply_argb, byte_buffer_to_buffer   #@UnresolvedImport
@@ -55,6 +55,7 @@ except:
     unpremultiply_argb, byte_buffer_to_buffer  = None, None
 
 
+
 class GTKKeyEvent(AdHocStruct):
     pass
 
@@ -65,21 +66,36 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         self._fullscreen = None
         self._iconified = False
         self._resize_counter = 0
-        self._window_workspace = self._client_properties.get("workspace")
+        self._window_workspace = self._client_properties.get("workspace", -1)
+        workspacelog("init_window(..) workspace=%s", self._window_workspace)
         self._desktop_workspace = -1
         ClientWindowBase.init_window(self, metadata)
         self._can_set_workspace = HAS_X11_BINDINGS and CAN_SET_WORKSPACE
 
 
     def setup_window(self):
-        ClientWindowBase.setup_window(self)
         self.set_app_paintable(True)
         self.add_events(self.WINDOW_EVENT_MASK)
+
+        #try to enable alpha on this window if needed,
+        #and if the backing class can support it:
+        bc = self.get_backing_class()
+        log("setup_window() has_alpha=%s, %s.HAS_ALPHA=%s", self._has_alpha, bc, bc.HAS_ALPHA)
+        if self._has_alpha and bc.HAS_ALPHA:
+            screen = self.get_screen()
+            rgba = screen.get_rgba_colormap()
+            if rgba is None:
+                log.error("cannot handle window transparency on screen %s", screen)
+            else:
+                self.set_colormap(rgba)
+                self._window_alpha = True
+
         if self._override_redirect:
             transient_for = self.get_transient_for()
             type_hint = self.get_type_hint()
             if transient_for is not None and transient_for.window is not None and type_hint in self.OR_TYPE_HINTS:
                 transient_for._override_redirect_windows.append(self)
+
         if not self._override_redirect:
             self.connect("notify::has-toplevel-focus", self._focus_change)
         def focus_in(*args):
@@ -92,6 +108,9 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.connect("property-notify-event", self.property_changed)
         self.connect("window-state-event", self.window_state_updated)
 
+        #this will create the backing:
+        ClientWindowBase.setup_window(self)
+
         self.move(*self._pos)
         self.set_default_size(*self._size)
 
@@ -168,10 +187,10 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.do_workspace_changed(event)
 
     def workspace_changed(self):
-        #on X11 clients, this fires from the root window property watcher 
+        #on X11 clients, this fires from the root window property watcher
         ClientWindowBase.workspace_changed(self)
         self.do_workspace_changed("desktop workspace changed")
-    
+
     def do_workspace_changed(self, info):
         #call this method whenever something workspace related may have changed
         window_workspace = self.get_window_workspace()
@@ -180,10 +199,6 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         if self._window_workspace==window_workspace and self._desktop_workspace==desktop_workspace:
             #no change
             return
-        if desktop_workspace<0 or window_workspace<0:
-            #not sure about which workspace we are on!?
-            #maybe the property has been cleared? maybe the window is being scrubbed? do nothing.
-            return
         if not self._client.window_refresh_config:
             workspacelog("sending configure event to update workspace value")
             self.process_configure_event()
@@ -191,9 +206,13 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #we can tell the server using a "buffer-refresh" packet instead
         #and also take care of tweaking the batch config
         client_properties = {"workspace" : window_workspace}
-        suspend_resume = None
         options = {"refresh-now" : False}               #no need to refresh it
-        if desktop_workspace!=window_workspace:
+        suspend_resume = None
+        if desktop_workspace<0 or window_workspace<0:
+            #maybe the property has been cleared? maybe the window is being scrubbed?
+            workspacelog("not sure if the window is shown or not: %s vs %s, resuming to be safe", desktop_workspace, window_workspace)
+            suspend_resume = False
+        elif desktop_workspace!=window_workspace:
             workspacelog("window is on a different workspace, increasing its batch delay")
             suspend_resume = True
         elif self._window_workspace!=self._desktop_workspace:
@@ -204,20 +223,26 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         self._desktop_workspace = desktop_workspace
 
 
+    def get_workspace_count(self):
+        if not HAS_X11_BINDINGS:
+            return 1
+        return self.xget_u32_property(root, "_NET_NUMBER_OF_DESKTOPS")
+
     def set_workspace(self):
-        if not HAS_X11_BINDINGS or not self._been_mapped:
+        if not self._can_set_workspace:
             return -1
         root = self.gdk_window().get_screen().get_root_window()
-        ndesktops = self.xget_u32_property(root, "_NET_NUMBER_OF_DESKTOPS")
+        ndesktops = self.get_workspace_count()
         workspacelog("%s.set_workspace() workspace=%s ndesktops=%s", self, self._window_workspace, ndesktops)
-        if ndesktops is None or ndesktops<=1:
+        if ndesktops is None or ndesktops<=1 or self._window_workspace<0:
             return  -1
         workspace = max(0, min(ndesktops-1, self._window_workspace))
         event_mask = SubstructureNotifyMask | SubstructureRedirectMask
 
         def send():
-            root_xid = root.xid
-            xwin = self.gdk_window().xid
+            from xpra.gtk_common.gobject_compat import get_xid
+            root_xid = get_xid(root)
+            xwin = get_xid(self.gdk_window())
             X11Window.sendClientMessage(root_xid, xwin, False, event_mask, "_NET_WM_DESKTOP",
                   workspace, CurrentTime, 0, 0, 0)
         trap.call_synced(send)
@@ -245,30 +270,6 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             if window:
                 self.set_transient_for(window)
 
-    def update_icon(self, width, height, coding, data):
-        log("%s.update_icon(%s, %s, %s, %s bytes)", self, width, height, coding, len(data))
-        if coding == "premult_argb32":
-            if unpremultiply_argb is not None:
-                #we usually cannot do in-place and this is not performance critical
-                data = byte_buffer_to_buffer(unpremultiply_argb(data))
-                pixbuf = gdk.pixbuf_new_from_data(data, gtk.gdk.COLORSPACE_RGB, True, 8, width, height, width*4)
-            else:
-                # slower fallback: we round-trip through PNG.
-                # This is ridiculous, but faster than doing a bunch of alpha
-                # un-premultiplying and byte-swapping by hand in Python
-                cairo_surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
-                cairo_surf.get_data()[:] = data
-                loader = gdk.PixbufLoader()
-                cairo_surf.write_to_png(loader)
-                loader.close()
-                pixbuf = loader.get_pixbuf()
-        else:
-            loader = gdk.PixbufLoader(coding)
-            loader.write(data, len(data))
-            loader.close()
-            pixbuf = loader.get_pixbuf()
-        self.set_icon(pixbuf)
-
 
     def paint_spinner(self, context, area):
         log("%s.paint_spinner(%s, %s)", self, context, area)
@@ -306,7 +307,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
     def do_map_event(self, event):
         log("%s.do_map_event(%s) OR=%s", self, event, self._override_redirect)
         gtk.Window.do_map_event(self, event)
-        xid = self._metadata.get("xid")
+        xid = self._metadata.strget("xid")
         if xid:
             self.set_xid(xid)
         if not self._override_redirect:
@@ -318,6 +319,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         props = self._client_properties
         self._client_properties = {}
         if not self._been_mapped:
+            #this is the first time around, so set the workspace:
             workspace = self.set_workspace()
         else:
             #window has been mapped, so these attributes can be read (if present):
@@ -326,7 +328,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             if workspace<0:
                 workspace = self.get_desktop_workspace()
         if self._window_workspace!=workspace:
-            workspacelog("map event: changed workspace from %s to %s", self._window_workspace, workspace)
+            workspacelog("map event: been_mapped=%s, changed workspace from %s to %s", self._been_mapped, self._window_workspace, workspace)
             self._window_workspace = workspace
             props["workspace"] = workspace
         log("map-window for wid=%s with client props=%s", self._id, props)
@@ -382,17 +384,18 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         w = max(1, w)
         h = max(1, h)
         self._resize_counter = resize_counter
-        if self.window.get_position()==(x, y):
+        window = self.gdk_window()
+        if window.get_position()==(x, y):
             #same location, just resize:
             if self._size!=(w, h):
-                self.window.resize(w, h)
+                self.resize(w, h)
         else:
-            mw, mh = gtk.gdk.get_default_root_window().get_size()
+            mw, mh = self._client.get_root_size()
             if not self.is_realized():
                 self.realize()
             #adjust for window frame:
-            ox, oy = self.window.get_origin()
-            rx, ry = self.window.get_root_origin()
+            ox, oy = window.get_origin()[-2:]
+            rx, ry = window.get_root_origin()
             ax = x - (ox - rx)
             ay = y - (oy - ry)
             #validate against edge of screen (ensure window is shown):
@@ -405,10 +408,10 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             elif ay >= mh:
                 ay = mh -1
             if self._size!=(w, h):
-                self.window.move_resize(ax, ay, w, h)
+                window.move_resize(ax, ay, w, h)
             else:
                 #just move:
-                self.window.move(ax, ay)
+                window.move(ax, ay)
         if self._size!=(w, h):
             self._size = (w, h)
             self.new_backing(w, h)
@@ -433,11 +436,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         pointer = (int(event.x_root), int(event.y_root))
         modifiers = self._client.mask_to_names(event.state)
         buttons = []
-        for mask, button in {gtk.gdk.BUTTON1_MASK : 1,
-                             gtk.gdk.BUTTON2_MASK : 2,
-                             gtk.gdk.BUTTON3_MASK : 3,
-                             gtk.gdk.BUTTON4_MASK : 4,
-                             gtk.gdk.BUTTON5_MASK : 5}.items():
+        for mask, button in self.BUTTON_MASK.items():
             if event.state & mask:
                 buttons.append(button)
         return pointer, modifiers, buttons
@@ -469,3 +468,23 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         focuslog("%s focus_change(%s) has-toplevel-focus=%s, _been_mapped=%s", self, args, htf, self._been_mapped)
         if self._been_mapped:
             self._client.update_focus(self._id, htf)
+
+
+    def update_icon(self, width, height, coding, data):
+        log("%s.update_icon(%s, %s, %s, %s bytes)", self, width, height, coding, len(data))
+        coding = bytestostr(coding)
+        if coding == "premult_argb32":
+            if unpremultiply_argb is None:
+                #we could use PIL here with mode 'RGBa'
+                log.warn("cannot process premult_argb32 icon without the argb module")
+                return
+            #we usually cannot do in-place and this is not performance critical
+            data = byte_buffer_to_buffer(unpremultiply_argb(data))
+            pixbuf = pixbuf_new_from_data(data, COLORSPACE_RGB, True, 8, width, height, width*4)
+        else:
+            loader = PixbufLoader()
+            loader.write(data)
+            loader.close()
+            pixbuf = loader.get_pixbuf()
+        log("%s.set_icon(%s)", self, pixbuf)
+        self.set_icon(pixbuf)
