@@ -1,16 +1,31 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 # Platform-specific code for Win32.
 
+import errno
 import os.path
 import sys
 
 #redirect output if we are launched from py2exe's gui mode:
 REDIRECT_OUTPUT = hasattr(sys, "frozen") and sys.frozen=="windows_exe"
+if getattr(sys, 'frozen', False):
+    #cx_freeze paths:
+    def jedir(relpathname):
+        return os.path.join(os.path.dirname(sys.executable), relpathname)
+    def addsyspath(relpathname):
+        v = jedir(relpathname)
+        if os.path.exists(v):
+            sys.path.append(v)
+    addsyspath('bin')
+    addsyspath('bin\\etc')
+    addsyspath('bin\\lib')
+    addsyspath('bin\\share')
+    addsyspath('bin\\library.zip')
+    os.environ['GI_TYPELIB_PATH'] = jedir('bin\\lib\girepository-1.0')
 
 
 def set_prgname(name):
@@ -153,7 +168,8 @@ def fix_unicode_out():
                                 if not remaining:
                                     break
                                 text = text[n.value:]
-                    except Exception as e:
+                    except:
+                        e = sys.exc_info()[1]
                         _complain("%s.write: %r" % (self.name, e))
                         raise
 
@@ -161,7 +177,8 @@ def fix_unicode_out():
                     try:
                         for line in lines:
                             self.write(line)
-                    except Exception as e:
+                    except:
+                        e = sys.exc_info()[1]
                         _complain("%s.writelines: %r" % (self.name, e))
                         raise
 
@@ -181,7 +198,9 @@ _wait_for_input = False
 
 def do_init():
     if not REDIRECT_OUTPUT:
-        fix_unicode_out()
+        if sys.version_info[0]<3:
+            #don't know why this breaks with Python 3 yet...
+            fix_unicode_out()
         #figure out if we want to wait for input at the end:
         global _wait_for_input
         try:
@@ -192,8 +211,31 @@ def do_init():
             cpos = console_info["CursorPosition"]
             #wait for input if this is a brand new console:
             _wait_for_input = cpos.X==0 and cpos.Y==0
-        except Exception, e:
-            print("error accessing console: %s" % e)
+        except:
+            e = sys.exc_info()[1]
+            code = -1
+            try:
+                code = e.winerror
+            except:
+                pass
+            if code==errno.ENXIO:
+                #ignore "no device" errors silently
+                #(ie: happens if you redirect the command to a file)
+                #we could also re-use the code above from "not_a_console()"
+                pass
+            else:
+                print("error accessing console %s: %s" % (errno.errorcode.get(e.errno, e.errno), e))
+        if _wait_for_input:
+            #don't wait for input when running under wine
+            #(which usually does not popup a new shell window)
+            try:
+                import win32api     #@UnresolvedImport
+                hKey = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE, r"Software\\Wine")
+                if hKey is not None:
+                    _wait_for_input = False
+            except:
+                #no wine key, assume not present and wait for input
+                pass
         return
     from xpra.platform import get_prgname
     LOG_FILENAME = get_prgname()+".log"
