@@ -5,23 +5,21 @@
 # later version. See the file COPYING for details.
 
 import time
+from math import sqrt
 
 from xpra.log import Logger
 log = Logger("window", "encoding")
 
 from xpra.net.protocol import compressed_wrapper, Compressed, use_lz4
-try:
-    from xpra.codecs.xor import xor_str        #@UnresolvedImport
-except Exception, e:
-    log("cannot load xor module: %s", e)
-    xor_str = None
 try:
     from xpra.codecs.argb.argb import bgra_to_rgb, bgra_to_rgba, argb_to_rgb, argb_to_rgba   #@UnresolvedImport
 except Exception, e:
     log("cannot load argb module: %s", e)
     bgra_to_rgb, bgra_to_rgba, argb_to_rgb, argb_to_rgba = (None,)*4
 from xpra.os_util import StringIOClass
-from xpra.codecs.loader import get_codec, has_codec
+from xpra.codecs.loader import get_codec, get_codec_version, has_codec
+from xpra.os_util import builtins
+_memoryview = builtins.__dict__.get("memoryview")
 try:
     from xpra.net.mmap_pipe import mmap_write
 except:
@@ -29,9 +27,10 @@ except:
 
 
 PIL = get_codec("PIL")
-enc_webp = get_codec("enc_webp")
-webp_handlers = get_codec("webp_bitmap_handlers")
-
+PIL_VERSION = get_codec_version("PIL")
+enc_webm = get_codec("enc_webm")
+webp_handlers = get_codec("webm_bitmap_handlers")
+PIL_can_optimize = PIL_VERSION>="2.2"
 
 #give warning message just once per key then ignore:
 encoding_warnings = set()
@@ -42,8 +41,39 @@ def warn_encoding_once(key, message):
         encoding_warnings.add(key)
 
 
-def webp_encode(coding, image, quality):
-    assert enc_webp and webp_handlers, "webp components are missing"
+def webp_encode(coding, image, supports_transparency, quality, speed, options):
+    stride = image.get_rowstride()
+    enc_webp = get_codec("enc_webp")
+    if enc_webp and stride>0 and stride%4==0 and image.get_pixel_format() in ("BGRA", "BGRX"):
+        #prefer Cython module:
+        alpha = supports_transparency and image.get_pixel_format().find("A")>=0
+        w = image.get_width()
+        h = image.get_height()
+        if quality==100:
+            #webp lossless is unbearibly slow for only marginal compression improvements,
+            #so force max speed:
+            speed = 100
+        else:
+            #normalize speed for webp: avoid low speeds!
+            speed = int(sqrt(speed) * 10)
+        speed = max(0, min(100, speed))
+        cdata = enc_webp.compress(image.get_pixels(), w, h, stride=stride/4, quality=quality, speed=speed, has_alpha=alpha)
+        client_options = {"speed" : speed}
+        if quality>=0 and quality<100:
+            client_options["quality"] = quality
+        if alpha:
+            client_options["has_alpha"] = True
+        return "webp", Compressed("webp", cdata), client_options, image.get_width(), image.get_height(), 0, 24
+    enc_webm = get_codec("enc_webm")
+    webp_handlers = get_codec("webm_bitmap_handlers")
+    if enc_webm and webp_handlers:
+        return webm_encode(image, quality)
+    #fallback to PIL
+    return PIL_encode(coding, image, quality, speed, supports_transparency)
+
+
+def webm_encode(image, quality):
+    assert enc_webm and webp_handlers, "webp components are missing"
 
     BitmapHandler = webp_handlers.BitmapHandler
     handler_encs = {
@@ -60,16 +90,16 @@ def webp_encode(coding, image, quality):
     bh, lossy_enc, lossless_enc, has_alpha = h_e
     q = max(1, quality)
     enc = None
-    if q>=100 and has_codec("enc_webp_lossless"):
-        enc = getattr(enc_webp, lossless_enc)
+    if q>=100 and has_codec("enc_webm_lossless"):
+        enc = getattr(enc_webm, lossless_enc)
         kwargs = {}
         client_options = {}
-        log("webp_encode(%s, %s) using lossless encoder=%s for %s", image, enc, pixel_format)
+        log("webm_encode(%s, %s) using lossless encoder=%s for %s", image, enc, pixel_format)
     if enc is None:
-        enc = getattr(enc_webp, lossy_enc)
+        enc = getattr(enc_webm, lossy_enc)
         kwargs = {"quality" : q}
         client_options = {"quality" : q}
-        log("webp_encode(%s, %s) using lossy encoder=%s with quality=%s for %s", image, enc, q, pixel_format)
+        log("webm_encode(%s, %s) using lossy encoder=%s with quality=%s for %s", image, enc, q, pixel_format)
     handler = BitmapHandler(image.get_pixels(), bh, image.get_width(), image.get_height(), image.get_rowstride())
     bpp = 24
     if has_alpha:
@@ -78,7 +108,10 @@ def webp_encode(coding, image, quality):
     return "webp", Compressed("webp", str(enc(handler, **kwargs).data)), client_options, image.get_width(), image.get_height(), 0, bpp
 
 
-def rgb_encode(coding, image, rgb_formats, supports_transparency, speed, rgb_zlib, rgb_lz4, encoding_client_options, supports_rgb24zlib):
+def roundup(n, m):
+    return (n + m - 1) & ~(m - 1)
+
+def rgb_encode(coding, image, rgb_formats, supports_transparency, speed, rgb_zlib=True, rgb_lz4=True, encoding_client_options=True, supports_rgb24zlib=True):
     pixel_format = image.get_pixel_format()
     #log("rgb_encode%s pixel_format=%s, rgb_formats=%s", (coding, image, rgb_formats, supports_transparency, speed, rgb_zlib, rgb_lz4, encoding_client_options, supports_rgb24zlib), pixel_format, rgb_formats)
     if pixel_format not in rgb_formats:
@@ -96,6 +129,27 @@ def rgb_encode(coding, image, rgb_formats, supports_transparency, speed, rgb_zli
     #compress here and return a wrapper so network code knows it is already zlib compressed:
     pixels = image.get_pixels()
 
+    #special case for when rowstride is so much bigger than the width
+    #that we would end up sending large chunks of padding with each row of pixels
+    #this happens with XShm pixel data (the default)
+    stride = image.get_rowstride()
+    width = image.get_width()
+    rstride = roundup(width*len(pixel_format), 4)   #a reasonable stride: rounded up to 4
+    height = image.get_height()
+    if stride>8 and rstride<stride:
+        al = len(pixels)                    #current buffer size
+        el = rstride*height                 #desirable size we could have
+        if al-el>1024 and el*110/100<al:    #is it worth re-striding to save space?
+            #we'll save at least 1KB and 10%, do it
+            #Note: we could also change the pixel format whilst we're at it
+            # and convert BGRX to RGB for example (assuming RGB is also supported by the client)
+            rows = []
+            for y in range(height):
+                rows.append(pixels[stride*y:stride*y+rstride])
+            pixels = "".join(rows)
+            log("rgb_encode: %s pixels re-stride saving %i%% from %s (%s bytes) to %s (%s bytes)", pixel_format, 100-100*el/al, stride, al, rstride, el)
+            stride = rstride
+
     #compression
     #by default, wire=raw:
     raw_data = str(pixels)
@@ -131,13 +185,13 @@ def rgb_encode(coding, image, rgb_formats, supports_transparency, speed, rgb_zli
         bpp = 24
     log("rgb_encode using level=%s, %s compressed %sx%s in %s/%s: %s bytes down to %s", level, algo, image.get_width(), image.get_height(), coding, pixel_format, len(pixels), len(raw_data))
     if not encoding_client_options or not supports_rgb24zlib:
-        return  coding, wire_data, {}, image.get_width(), image.get_height(), image.get_rowstride(), bpp
+        return  coding, wire_data, {}, width, height, stride, bpp
     #wrap it using "Compressed" so the network layer receiving it
     #won't decompress it (leave it to the client's draw thread)
-    return coding, Compressed(coding, raw_data), options, image.get_width(), image.get_height(), image.get_rowstride(), bpp
+    return coding, Compressed(coding, raw_data), options, width, height, stride, bpp
 
 
-def PIL_encode(coding, image, quality, supports_transparency):
+def PIL_encode(coding, image, quality, speed, supports_transparency):
     assert PIL is not None, "Python PIL is not available"
     pixel_format = image.get_pixel_format()
     w = image.get_width()
@@ -154,7 +208,11 @@ def PIL_encode(coding, image, quality, supports_transparency):
         #it is safe to use frombuffer() here since the convert()
         #calls below will not convert and modify the data in place
         #and we save the compressed data then discard the image
-        im = PIL.Image.frombuffer(rgb, (w, h), image.get_pixels(), "raw", pixel_format, image.get_rowstride())
+        pixels = image.get_pixels()
+        if _memoryview and isinstance(pixels, _memoryview):
+            #PIL cannot use the memoryview directly:
+            pixels = pixels.tobytes()
+        im = PIL.Image.frombuffer(rgb, (w, h), pixels, "raw", pixel_format, image.get_rowstride())
         if coding.startswith("png") and not supports_transparency and rgb=="RGBA":
             im = im.convert("RGB")
             rgb = "RGB"
@@ -164,11 +222,15 @@ def PIL_encode(coding, image, quality, supports_transparency):
         raise e
     buf = StringIOClass()
     client_options = {}
-    if coding=="jpeg":
+    #only optimize with Pillow>=2.2 and when speed is zero
+    if coding in ("jpeg", "webp"):
         q = int(min(99, max(1, quality)))
         kwargs = im.info
         kwargs["quality"] = q
-        im.save(buf, "JPEG", **kwargs)
+        if coding=="jpeg" and PIL_can_optimize and speed<70:
+            #(optimizing jpeg is pretty cheap and worth doing)
+            kwargs["optimize"] = True
+        im.save(buf, coding.upper(), **kwargs)
         client_options["quality"] = q
     else:
         assert coding in ("png", "png/P", "png/L"), "unsupported png encoding: %s" % coding
@@ -203,6 +265,17 @@ def PIL_encode(coding, image, quality, supports_transparency):
         if mask is not None:
             client_options["transparency"] = 255
             kwargs["transparency"] = 255
+        if PIL_can_optimize and speed==0:
+            #optimizing png is very rarely worth doing
+            kwargs["optimize"] = True
+        #level can range from 0 to 9, but anything above 5 is way too slow for small gains:
+        #76-100   -> 1
+        #51-76    -> 2
+        #etc
+        kwargs["compress_level"] = max(1, min(5, (125-speed)/25))
+        #default is good enough, no need to override, other options:
+        #DEFAULT_STRATEGY, FILTERED, HUFFMAN_ONLY, RLE, FIXED
+        #kwargs["compress_type"] = PIL.Image.DEFAULT_STRATEGY
         im.save(buf, "PNG", **kwargs)
     log("sending %sx%s %s as %s, mode=%s, options=%s", w, h, pixel_format, coding, im.mode, kwargs)
     data = buf.getvalue()
@@ -288,6 +361,9 @@ def rgb_reformat(image, rgb_formats, supports_transparency):
     start = time.time()
     w = image.get_width()
     h = image.get_height()
+    if _memoryview and isinstance(pixels, _memoryview):
+        #PIL cannot use the memoryview directly:
+        pixels = pixels.tobytes()
     img = PIL.Image.frombuffer(target_format, (w, h), pixels, "raw", input_format, image.get_rowstride())
     rowstride = w*len(target_format)    #number of characters is number of bytes per pixel!
     data = img.tostring("raw", target_format)
