@@ -23,6 +23,7 @@ from xpra.codecs.image_wrapper import ImageWrapper
 from xpra.codecs.video_helper import getVideoHelper, PREFERRED_ENCODER_ORDER
 from xpra.os_util import Queue, SIGNAMES
 from xpra.util import typedict
+from xpra.version_util import local_version
 from xpra.daemon_thread import make_daemon_thread
 from xpra.scripts.config import parse_number, parse_bool
 from xpra.scripts.server import create_unix_domain_socket
@@ -315,9 +316,9 @@ class ProxyInstanceProcess(Process):
 
 
     def get_proxy_info(self, proto):
-        info = {}
-        info.update(get_server_info("proxy."))
-        info.update(get_thread_info("proxy.", proto))
+        info = {"proxy.version" : local_version}
+        info.update(get_server_info("proxy"))
+        info.update(get_thread_info("proxy", proto))
         info.update(self.get_encoder_info())
         return info
 
@@ -368,7 +369,7 @@ class ProxyInstanceProcess(Process):
         #replace the network caps with the proxy's own:
         pcaps.update(get_network_caps())
         #then add the proxy info:
-        pcaps.update(get_server_info("proxy."))
+        pcaps.update(get_server_info("proxy"))
         pcaps["proxy"] = True
         pcaps["proxy.hostname"] = socket.gethostname()
         return pcaps
@@ -622,7 +623,7 @@ class ProxyInstanceProcess(Process):
                     passthrough()
                     return
                 log("no video encoder available: sending as jpeg")
-                coding, compressed_data, client_options, _, _, _, _ = PIL_encode("jpeg", image, quality, False)
+                coding, compressed_data, client_options, _, _, _, _ = PIL_encode("jpeg", image, quality, speed, False)
                 return send_updated(coding, compressed_data, client_options)
 
             log("creating new video encoder %s for window %s", spec, wid)
