@@ -19,14 +19,15 @@ from xpra.keyboard.mask import DEFAULT_MODIFIER_MEANINGS
 from xpra.server.server_core import ServerCore
 from xpra.util import log_screen_sizes
 from xpra.os_util import thread, get_hex_uuid
-from xpra.version_util import add_version_info
 from xpra.util import typedict
-from xpra.codecs.loader import PREFERED_ENCODING_ORDER, codec_versions, has_codec
+from xpra.codecs.loader import PREFERED_ENCODING_ORDER, codec_versions, has_codec, get_codec
+from xpra.codecs.codec_constants import get_PIL_encodings
 from xpra.codecs.video_helper import getVideoHelper, ALL_VIDEO_ENCODER_OPTIONS, get_DEFAULT_VIDEO_ENCODERS, ALL_CSC_MODULE_OPTIONS, get_DEFAULT_CSC_MODULES
 if sys.version > '3':
     unicode = str           #@ReservedAssignment
 
 
+DETECT_LEAKS = os.environ.get("XPRA_DETECT_LEAKS", "0")=="1"
 MAX_CONCURRENT_CONNECTIONS = 20
 
 
@@ -81,6 +82,7 @@ class ServerBase(ServerCore):
         #control mode:
         self.control_commands = ["hello", "help",
                     "debug",
+                    "encoding",
                     "quality", "min-quality", "speed", "min-speed",
                     "compression", "encoder", "refresh",
                     "sound-output",
@@ -92,6 +94,18 @@ class ServerBase(ServerCore):
         self.init_packet_handlers()
         self.init_aliases()
 
+        if DETECT_LEAKS:
+            from xpra.util import detect_leaks
+            detailed = []
+            #example: warning, uses ugly direct import:
+            #try:
+            #    from xpra.x11.bindings.ximage import XShmImageWrapper       #@UnresolvedImport
+            #    detailed.append(XShmImageWrapper)
+            #except:
+            #    pass
+            print_leaks = detect_leaks(log, detailed)
+            self.timeout_add(10*1000, print_leaks)
+
     def idle_add(self, *args, **kwargs):
         raise NotImplementedError()
 
@@ -119,6 +133,7 @@ class ServerBase(ServerCore):
         self.dpi = self.default_dpi
         self.supports_clipboard = opts.clipboard
         self.supports_dbus_proxy = opts.dbus_proxy
+        self.send_pings = opts.pings
 
         log("starting component init")
         self.init_clipboard(self.supports_clipboard, opts.clipboard_filter_file)
@@ -133,11 +148,6 @@ class ServerBase(ServerCore):
         getVideoHelper().set_modules(video_encoders=video_encoders, csc_modules=csc_modules)
 
         self.load_existing_windows(opts.system_tray)
-
-        if opts.pings:
-            self.timeout_add(1000, self.send_ping)
-        else:
-            self.timeout_add(10*1000, self.send_ping)
         thread.start_new_thread(self.threaded_init, ())
 
     def threaded_init(self):
@@ -149,33 +159,29 @@ class ServerBase(ServerCore):
         log("threaded_init() end")
 
     def init_encodings(self):
-        #core encodings: all the specific encoding formats we can encode:
-        self.core_encodings = ["rgb24", "rgb32"]
-        #encodings: the format families we can encode (same as core, except for rgb):
-        self.encodings = ["rgb"]
-
         def add_encodings(encodings):
-            for e in encodings:
+            for ce in encodings:
+                e = {"rgb32" : "rgb", "rgb24" : "rgb"}.get(ce, ce)
                 if e not in self.encodings:
                     self.encodings.append(e)
-                if e not in self.core_encodings:
-                    self.core_encodings.append(e)
+                if ce not in self.core_encodings:
+                    self.core_encodings.append(ce)
+
+        add_encodings(["rgb24", "rgb32"])
 
         #video encoders (empty when first called - see threaded_init)
         add_encodings(getVideoHelper().get_encodings())  #ie: ["vp8", "h264"]
-
-        for module, encodings in {
-                              "enc_webp"  : ["webp"],
-                              "PIL"       : ["png", "png/L", "png/P", "jpeg"],
-                              }.items():
-            if not has_codec(module):
-                log("init_encodings() codec module %s is missing, not adding: %s", module, encodings)
-                continue
-            add_encodings(encodings)
+        #Pithon Imaging Libary:
+        PIL = get_codec("PIL")
+        if PIL:
+            add_encodings(get_PIL_encodings(PIL))
+        #webp: only check for "enc_webm" because "enc_webp" needs a fallback (either "PIL" or "enc_webm")
+        if has_codec("enc_webm"):
+            add_encodings(["webp"])
 
         self.lossless_encodings = [x for x in self.core_encodings if (x.startswith("png") or x.startswith("rgb"))]
         self.lossless_mode_encodings = []
-        if has_codec("enc_webp_lossless"):
+        if has_codec("enc_webm_lossless"):
             self.lossless_mode_encodings.append("webp")
             self.lossless_encodings.append("webp")
 
@@ -372,6 +378,14 @@ class ServerBase(ServerCore):
         self.do_init_aliases(packet_types)
 
 
+    def run(self):
+        if self.send_pings:
+            self.timeout_add(1000, self.send_ping)
+        else:
+            self.timeout_add(10*1000, self.send_ping)
+        return ServerCore.run(self)
+
+
     def cleanup(self, *args):
         if self.notifications_forwarder:
             thread.start_new_thread(self.notifications_forwarder.release, ())
@@ -503,15 +517,21 @@ class ServerBase(ServerCore):
                           self.default_speed, self.default_min_speed)
         log("process_hello serversource=%s", ss)
         ss.parse_hello(c)
-        try:
-            dw, dh = ss.desktop_size
-            log.info("client root window size is %sx%s with %s displays:", dw, dh, len(ss.screen_sizes))
-            log_screen_sizes(dw, dh, ss.screen_sizes)
-        except:
-            pass
+        dw, dh = None, None
+        if ss.desktop_size:
+            try:
+                dw, dh = ss.desktop_size
+                if not ss.screen_sizes:
+                    log.info("client root window size is %sx%s", dw, dh)
+                else:
+                    log.info("client root window size is %sx%s with %s displays:", dw, dh, len(ss.screen_sizes))
+                    log_screen_sizes(dw, dh, ss.screen_sizes)
+            except:
+                dw, dh = None, None
         self._server_sources[proto] = ss
         root_w, root_h = self.set_best_screen_size()
         self.calculate_workarea()
+        self.set_desktop_geometry(dw or root_w, dh or root_h)
         #take the clipboard if no-one else has yet:
         if ss.clipboard_enabled and self._clipboard_helper is not None and \
             (self._clipboard_client is None or self._clipboard_client.is_closed()):
@@ -532,6 +552,9 @@ class ServerBase(ServerCore):
             log("process_hello server has clipboards: %s, client supports: %s", self._clipboards, client_selections)
             self._clipboard_helper.enable_selections(client_selections)
 
+        #keyboard:
+        ss.keyboard_config = self.get_keyboard_config(c)
+
         #so only activate this feature afterwards:
         self.keyboard_sync = c.boolget("keyboard_sync", True)
         key_repeat = c.intpair("key_repeat")
@@ -552,6 +575,10 @@ class ServerBase(ServerCore):
     def update_server_settings(self, settings, reset=False):
         log("server settings ignored: ", settings)
 
+
+    def get_keyboard_config(self, props):
+        return None
+
     def set_keyboard_repeat(self, key_repeat):
         pass
 
@@ -601,7 +628,6 @@ class ServerBase(ServerCore):
              "sound.server_driven"          : True,
              "server_type"                  : "base",
              })
-        add_version_info(capabilities)
         for k,v in codec_versions.items():
             capabilities["encoding.%s.version" % k] = v
         return capabilities
@@ -616,8 +642,8 @@ class ServerBase(ServerCore):
              "encodings"                : self.encodings,
              "encodings.core"           : self.core_encodings,
              "encodings.lossless"       : self.lossless_encodings,
-             "encodings.with_speed"     : [x for x in self.core_encodings if x in ("h264", "rgb")],
-             "encodings.with_quality"   : [x for x in self.core_encodings if x in ("jpeg", "webp", "h264")],
+             "encodings.with_speed"     : [x for x in self.core_encodings if x in ("h264", "vp8", "vp9", "rgb")],
+             "encodings.with_quality"   : [x for x in self.core_encodings if x in ("jpeg", "webp", "h264", "vp8", "vp9")],
              "encodings.with_lossless_mode" : self.lossless_mode_encodings}
 
     def send_hello(self, server_source, root_w, root_h, key_repeat, server_cipher):
@@ -769,6 +795,30 @@ class ServerBase(ServerCore):
             for csource in sources:
                 csource.pointer_ungrab(-1)
             return respond(0, "ungrabbed %s clients" % len(sources))
+        elif command=="encoding":
+            if len(args)<1:
+                return argn_err(1)
+            encoding = args[0]
+            args = args[1:]
+            if len(args)>0:
+                wids = []
+                for x in args:
+                    try:
+                        wid = int(x)
+                        wids.append(wid)
+                    except:
+                        return respond(4, "invalid window id %s" % x)
+            else:
+                wids = self._id_to_window.keys()
+            for csource in sources:
+                csource.set_encoding(encoding, wids)
+            #now also do a refresh:
+            for wid in wids:
+                window = self._id_to_window.get(wid)
+                if window:
+                    for csource in sources:
+                        csource.refresh(wid, window, {})
+            return respond(0, "set encoding to %s for %s windows" % (encoding, len(wids)))
         elif command=="refresh":
             if len(args)>0:
                 widwin = []
@@ -820,6 +870,10 @@ class ServerBase(ServerCore):
                     assert False, "invalid command: %s" % command
             wid_str = args[0]
             for_all_window_sources(wid_str, set_value)
+            #update the default encoding options 
+            #so new windows will also inherit those settings: 
+            for csource in sources: 
+                csource.default_encoding_options[command] = v 
             return respond(0, "%s set to %s on window %s for %s clients" % (command, v, wid_str, len(sources)))
         elif command=="client":
             if len(args)==0:
@@ -835,7 +889,7 @@ class ServerBase(ServerCore):
             csource.send_client_command(*client_command)
             return respond(0, "client control command '%s' forwarded to %s clients" % (client_command[0], count))
         else:
-            return respond(9, "internal state error: invalid command '%s'", command)
+            return respond(9, "internal state error: invalid command '%s'" % command)
 
 
     def send_screenshot(self, proto):
@@ -860,6 +914,8 @@ class ServerBase(ServerCore):
             (ie: things that query the display)
         """
         info = {"server.max_desktop_size" : self.get_max_screen_size()}
+        if self.keyboard_config:
+            info["state.modifiers"] = self.keyboard_config.get_current_mask()
         #window info:
         self.add_windows_info(info, wids)
         return info
@@ -1071,10 +1127,11 @@ class ServerBase(ServerCore):
         ss = self._server_sources.get(proto)
         if ss is None:
             return
+        if len(packet)>=4:
+            ss.set_screen_sizes(packet[3])
         log("client requesting new size: %sx%s", width, height)
         self.set_screen_size(width, height)
         if len(packet)>=4:
-            ss.set_screen_sizes(packet[3])
             log.info("received updated display dimensions")
             log.info("client root window size is %sx%s with %s displays:", width, height, len(ss.screen_sizes))
             log_screen_sizes(width, height, ss.screen_sizes)
@@ -1249,8 +1306,13 @@ class ServerBase(ServerCore):
         if ss:
             ss.set_client_properties(wid, window, typedict(new_client_properties))
             client_properties = self.client_properties.setdefault("%s|%s" % (wid, ss.uuid), {})
-            log("set_client_properties updating window %s with %s", wid, new_client_properties)
-            client_properties.update(new_client_properties)
+            #filter out encoding properties, which are expected to be set everytime:
+            ncp = {}
+            for k,v in new_client_properties.items():
+                if not k.startswith("encoding"):
+                    ncp[k] = v
+            log("set_client_properties updating window %s with %s", wid, ncp)
+            client_properties.update(ncp)
 
 
     def _process_focus(self, proto, packet):
@@ -1271,12 +1333,14 @@ class ServerBase(ServerCore):
             self.set_keymap(ss, force=True)
 
     def _process_keymap(self, proto, packet):
-        props = packet[1]
+        props = typedict(packet[1])
         ss = self._server_sources.get(proto)
         if ss is None:
             return
         log("received new keymap from client")
-        if ss.assign_keymap_options(props):
+        kc = ss.keyboard_config
+        if kc and kc.enabled:
+            kc.parse_options(props)
             self.set_keymap(ss, True)
         modifiers = props.get("modifiers")
         ss.make_keymask_match(modifiers)
@@ -1301,7 +1365,12 @@ class ServerBase(ServerCore):
         return ss.get_keycode(client_keycode, keyname, modifiers)
 
     def is_modifier(self, keyname, keycode):
-        return keyname in DEFAULT_MODIFIER_MEANINGS.keys()
+        if keyname in DEFAULT_MODIFIER_MEANINGS.keys():
+            return True
+        #keyboard config should always exist if we are here?
+        if self.keyboard_config:
+            return self.keyboard_config.is_modifier(keycode)
+        return False
 
     def fake_key(self, keycode, press):
         pass
@@ -1322,20 +1391,21 @@ class ServerBase(ServerCore):
             del self.keys_timedout[keycode]
         def press():
             keylog("handle keycode pressing %s: key %s", keycode, name)
-            if self.keyboard_sync:
-                self.keys_pressed[keycode] = name
+            self.keys_pressed[keycode] = name
             self.fake_key(keycode, True)
         def unpress():
             keylog("handle keycode unpressing %s: key %s", keycode, name)
-            if self.keyboard_sync:
+            if keycode in self.keys_pressed:
                 del self.keys_pressed[keycode]
             self.fake_key(keycode, False)
+        is_mod = self.is_modifier(name, keycode)
         if pressed:
             if keycode not in self.keys_pressed:
                 press()
-                if not self.keyboard_sync:
+                if not self.keyboard_sync and not is_mod:
                     #keyboard is not synced: client manages repeat so unpress
-                    #it immediately
+                    #it immediately unless this is a modifier key
+                    #(as modifiers are synced via many packets: key, focus and mouse events)
                     unpress()
             else:
                 keylog("handle keycode %s: key %s was already pressed, ignoring", keycode, name)
@@ -1344,7 +1414,6 @@ class ServerBase(ServerCore):
                 unpress()
             else:
                 keylog("handle keycode %s: key %s was already unpressed, ignoring", keycode, name)
-        is_mod = self.is_modifier(name, keycode)
         if not is_mod and self.keyboard_sync and self.key_repeat_delay>0 and self.key_repeat_interval>0:
             self._key_repeat(wid, pressed, name, keyval, keycode, modifiers, self.key_repeat_delay)
 
@@ -1352,17 +1421,18 @@ class ServerBase(ServerCore):
         """ Schedules/cancels the key repeat timeouts """
         timer = self.keys_repeat_timers.get(keycode, None)
         if timer:
+            del self.keys_repeat_timers[keycode]
             keylog("cancelling key repeat timer: %s for %s / %s", timer, keyname, keycode)
             self.source_remove(timer)
         if pressed:
             delay_ms = min(1500, max(250, delay_ms))
             keylog("scheduling key repeat timer with delay %s for %s / %s", delay_ms, keyname, keycode)
             def _key_repeat_timeout(when):
+                del self.keys_repeat_timers[keycode]
                 now = time.time()
                 keylog("key repeat timeout for %s / '%s' - clearing it, now=%s, scheduled at %s with delay=%s", keyname, keycode, now, when, delay_ms)
                 self._handle_key(wid, False, keyname, keyval, keycode, modifiers)
                 self.keys_timedout[keycode] = now
-                del self.keys_repeat_timers[keycode]
             now = time.time()
             self.keys_repeat_timers[keycode] = self.timeout_add(delay_ms, _key_repeat_timeout, now)
 
