@@ -11,17 +11,14 @@ events, mucking about with properties, etc. etc.) is wrapped up in here."""
 
 # Maintain compatibility with old versions of Python, while avoiding a
 # deprecation warning on new versions:
-import sys
-if sys.version_info < (2, 6):
-    from sets import ImmutableSet
-else:
-    ImmutableSet = frozenset
-import gobject
-import gtk.gdk
-import cairo
 import os
 from socket import gethostname
 
+import gobject
+import gtk.gdk
+import cairo
+
+from xpra.util import nonl
 from xpra.x11.bindings.window_bindings import constants, X11WindowBindings #@UnresolvedImport
 X11Window = X11WindowBindings()
 
@@ -38,14 +35,15 @@ from xpra.x11.gtk_x11.send_wm import (
 from xpra.gtk_common.gobject_util import (AutoPropGObjectMixin,
                            one_arg_signal,
                            non_none_list_accumulator)
-from xpra.x11.gtk_x11.error import trap, XError
+from xpra.gtk_common.error import trap, XError
 from xpra.x11.gtk_x11.prop import prop_get, prop_set
 from xpra.x11.gtk_x11.composite import CompositeHelper
-from xpra.x11.gtk_x11.pointer_grab import PointerGrabHelper
+from xpra.os_util import ImmutableSet
 
 from xpra.log import Logger
 log = Logger("x11", "window")
 focuslog = Logger("x11", "window", "focus")
+grablog = Logger("x11", "window", "grab")
 
 if gtk.pygtk_version<(2,17):
     log.error("your version of PyGTK is too old - expect some bugs")
@@ -59,6 +57,25 @@ CWHeight = constants["CWHeight"]
 IconicState = constants["IconicState"]
 NormalState = constants["NormalState"]
 
+# grab stuff:
+NotifyNormal        = constants["NotifyNormal"]
+NotifyGrab          = constants["NotifyGrab"]
+NotifyUngrab        = constants["NotifyUngrab"]
+NotifyWhileGrabbed  = constants["NotifyWhileGrabbed"]
+GRAB_CONSTANTS = {
+                  NotifyNormal          : "NotifyNormal",
+                  NotifyGrab            : "NotifyGrab",
+                  NotifyUngrab          : "NotifyUngrab",
+                  NotifyWhileGrabbed    : "NotifyWhileGrabbed",
+                 }
+DETAIL_CONSTANTS    = {}
+for x in ("NotifyAncestor", "NotifyVirtual", "NotifyInferior",
+          "NotifyNonlinear", "NotifyNonlinearVirtual", "NotifyPointer",
+          "NotifyPointerRoot", "NotifyDetailNone"):
+    DETAIL_CONSTANTS[constants[x]] = x
+grablog("pointer grab constants: %s", GRAB_CONSTANTS)
+grablog("detail constants: %s", DETAIL_CONSTANTS)
+
 
 #if you want to use a virtual screen bigger than 32767x32767
 #you will need to change those values, but some broken toolkits
@@ -220,6 +237,10 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                 "PID of owning process", "",
                 -1, 65535, -1,
                 gobject.PARAM_READABLE),
+        "opacity": (gobject.TYPE_INT64,
+                "Opacity", "",
+                -1, 0xffffffff, -1,
+                gobject.PARAM_READABLE),
         "xid": (gobject.TYPE_INT,
                 "X11 window id", "",
                 -1, 65535, -1,
@@ -242,10 +263,6 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                        "Does the window use transparency", "",
                        False,
                        gobject.PARAM_READABLE),
-        "has-grab": (gobject.TYPE_BOOLEAN,
-                       "Does the window own a grab", "",
-                       False,
-                       gobject.PARAM_READABLE),
         "fullscreen": (gobject.TYPE_BOOLEAN,
                        "Fullscreen-ness of window", "",
                        False,
@@ -281,11 +298,12 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         "raised"                : one_arg_signal,
         "unmanaged"             : one_arg_signal,
 
-        "pointer-grab"          : one_arg_signal,
-        "pointer-ungrab"        : one_arg_signal,
-
-# this signal must be defined in the subclasses to be seen by the event stuff:
+        "grab"                  : one_arg_signal,
+        "ungrab"                : one_arg_signal,
+# these signals must be defined in the subclasses to be seen by the event stuff:
 #        "xpra-configure-event": one_arg_signal,
+#        "xpra-focus-in-event"   : one_arg_signal,
+#        "xpra-focus-out-event"  : one_arg_signal,
         }
 
     def __init__(self, client_window):
@@ -303,8 +321,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         self._internal_set_property("client-window", client_window)
         use_xshm = USE_XSHM and (not self.is_OR() and not self.is_tray())
         self._composite = CompositeHelper(self.client_window, False, use_xshm)
-        self._pointer_grab = PointerGrabHelper(self.client_window)
-        self.property_names = ["pid", "transient-for", "fullscreen", "maximized", "window-type", "role", "group-leader", "xid", "has-alpha"]
+        self.property_names = ["pid", "transient-for", "fullscreen", "maximized", "window-type", "role", "group-leader", "xid", "has-alpha", "opacity"]
 
     def get_property_names(self):
         return self.property_names
@@ -318,6 +335,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def managed_disconnect(self):
         for handler_id in self._managed_handlers:
             self.disconnect(handler_id)
+        self._managed_handlers = []
 
     def call_setup(self):
         try:
@@ -346,9 +364,6 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             except Exception, ex:
                 log.error("error in cleanup handler: %s", ex)
             raise Unmanageable(e)
-        self._pointer_grab.setup()
-        self._pointer_grab.connect("grab", self.pointer_grab_event)
-        self._pointer_grab.connect("ungrab", self.pointer_ungrab_event)
         self._setup_done = True
 
     def setup_failed(self, e):
@@ -356,6 +371,11 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         self.do_unmanaged(False)
 
     def setup(self):
+        # Start listening for important events.
+        self.client_window.set_events(self.client_window_saved_events
+                                      | gtk.gdk.STRUCTURE_MASK
+                                      | gtk.gdk.PROPERTY_CHANGE_MASK
+                                      | gtk.gdk.FOCUS_CHANGE_MASK)
         h = self._composite.connect("contents-changed", self._forward_contents_changed)
         self._damage_forward_handle = h
 
@@ -378,6 +398,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             self.emit("client-contents-changed", event)
 
     def acknowledge_changes(self):
+        assert self._composite, "composite window destroyed outside the UI thread?"
         self._composite.acknowledge_changes()
 
     ################################
@@ -446,9 +467,6 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                 self._damage_forward_handle = None
             self._composite.destroy()
             self._composite = None
-        if self._pointer_grab:
-            self._pointer_grab.destroy()
-            self._pointer_grab = None
 
     def _read_initial_properties(self):
         def pget(key, ptype, raise_xerrors=True):
@@ -462,12 +480,11 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             window_type = self._guess_window_type(transient_for)
             window_types = [gtk.gdk.atom_intern(window_type)]
         self._internal_set_property("window-type", window_types)
-        self._handle_scaling()
         self._internal_set_property("has-alpha", self.client_window.get_depth()==32)
         self._internal_set_property("xid", self.client_window.xid)
         self._internal_set_property("pid", pget("_NET_WM_PID", "u32") or -1)
         self._internal_set_property("role", pget("WM_WINDOW_ROLE", "latin1"))
-        for mutable in ["WM_NAME", "_NET_WM_NAME"]:
+        for mutable in ["WM_NAME", "_NET_WM_NAME", "_NET_WM_WINDOW_OPACITY", "_XPRA_SCALING"]:
             self._call_property_handler(mutable)
 
 
@@ -482,6 +499,11 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             self._internal_set_property("scaling", None)
     _property_handlers["_XPRA_SCALING"] = _handle_scaling
 
+    def _handle_opacity_change(self):
+        opacity = self.prop_get("_NET_WM_WINDOW_OPACITY", "u32", True) or -1
+        self._internal_set_property("opacity", opacity)
+    _property_handlers["_NET_WM_WINDOW_OPACITY"] = _handle_opacity_change
+
     def _handle_title_change(self):
         net_wm_name = self.prop_get("_NET_WM_NAME", "utf8", True)
         if net_wm_name is not None:
@@ -608,16 +630,24 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def set_active(self):
         prop_set(self.client_window.get_screen().get_root_window(), "_NET_ACTIVE_WINDOW", "u32", self.client_window.xid)
 
-    #bits to do with grabs: just re-emit the signal
-    def pointer_grab_event(self, gh, event):
-        log("pointer_grab_event(%s, %s) forwarding it", gh, event)
-        self._internal_set_property("has-grab", True)
-        self.emit("pointer-grab", event)
 
-    def pointer_ungrab_event(self, gh, event):
-        log("pointer_ungrab_event(%s, %s) forwarding it", gh, event)
-        self._internal_set_property("has-grab", False)
-        self.emit("pointer-ungrab", event)
+    def do_xpra_focus_in_event(self, event):
+        grablog("focus_in_event(%s) mode=%s, detail=%s",
+            event, GRAB_CONSTANTS.get(event.mode), DETAIL_CONSTANTS.get(event.detail, event.detail))
+        self.may_emit_grab(event)
+
+    def do_xpra_focus_out_event(self, event):
+        grablog("focus_out_event(%s) mode=%s, detail=%s",
+            event, GRAB_CONSTANTS.get(event.mode), DETAIL_CONSTANTS.get(event.detail, event.detail))
+        self.may_emit_grab(event)
+
+    def may_emit_grab(self, event):
+        if event.mode==NotifyGrab:
+            grablog("emitting grab on %s", self)
+            self.emit("grab", event)
+        if event.mode==NotifyUngrab:
+            grablog("emitting ungrab on %s", self)
+            self.emit("ungrab", event)
 
 
 gobject.type_register(BaseWindowModel)
@@ -632,6 +662,8 @@ class OverrideRedirectWindowModel(BaseWindowModel):
         "xpra-unmap-event": one_arg_signal,
         "xpra-client-message-event" : one_arg_signal,
         "xpra-property-notify-event": one_arg_signal,
+        "xpra-focus-in-event"   : one_arg_signal,
+        "xpra-focus-out-event"  : one_arg_signal,
         }
 
     def __init__(self, client_window):
@@ -644,8 +676,6 @@ class OverrideRedirectWindowModel(BaseWindowModel):
 
     def setup(self):
         BaseWindowModel.setup(self)
-        self.client_window.set_events(self.client_window_saved_events
-                                      | gtk.gdk.STRUCTURE_MASK)
         # So now if the window becomes unmapped in the future then we will
         # notice... but it might be unmapped already, and any event
         # already generated, and our request for that event is too late!
@@ -783,6 +813,8 @@ class WindowModel(BaseWindowModel):
                                non_none_list_accumulator),
         "xpra-property-notify-event": one_arg_signal,
         "xpra-configure-event": one_arg_signal,
+        "xpra-focus-in-event"   : one_arg_signal,
+        "xpra-focus-out-event"  : one_arg_signal,
 
         "child-map-request-event": one_arg_signal,
         "child-configure-request-event": one_arg_signal,
@@ -827,11 +859,6 @@ class WindowModel(BaseWindowModel):
         X11Window.substructureRedirect(self.corral_window.xid)
         add_event_receiver(self.corral_window, self)
 
-        # Start listening for important events.
-        self.client_window.set_events(self.client_window_saved_events
-                                      | gtk.gdk.STRUCTURE_MASK
-                                      | gtk.gdk.PROPERTY_CHANGE_MASK)
-
         # The child might already be mapped, in case we inherited it from
         # a previous window manager.  If so, we unmap it now, and save the
         # serial number of the request -- this way, when we get an
@@ -1457,7 +1484,13 @@ class WindowModel(BaseWindowModel):
         trap.swallow_synced(X11Window.XKillClient, self.client_window.xid)
 
     def __repr__(self):
-        return "WindowModel(%#x)" % self.client_window.xid
+        xid = 0
+        if self.client_window:
+            xid = self.client_window.xid
+        title = self.get_property("title")
+        if title:
+            return "WindowModel(%#x - \"%s\")" % (xid, nonl(title))
+        return "WindowModel(%#x)" % xid
 
 
 gobject.type_register(WindowModel)
