@@ -32,11 +32,12 @@ X11Window = X11WindowBindings()
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
 X11Keyboard = X11KeyboardBindings()
 from xpra.x11.gtk_x11.window import OverrideRedirectWindowModel, SystemTrayWindowModel, Unmanageable
-from xpra.x11.gtk_x11.error import trap
+from xpra.gtk_common.error import trap
 
 from xpra.log import Logger
 log = Logger("server")
 focuslog = Logger("server", "focus")
+grablog = Logger("server", "grab")
 windowlog = Logger("server", "window")
 cursorlog = Logger("server", "cursor")
 traylog = Logger("server", "tray")
@@ -179,6 +180,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         X11ServerBase.x11_init(self)
         assert init_x11_filter() is True
 
+        self._has_grab = 0
         self._has_focus = 0
         # Do this before creating the Wm object, to avoid clobbering its
         # selecting SubstructureRedirect.
@@ -256,13 +258,21 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
     def get_window_info(self, window):
         info = X11ServerBase.get_window_info(self, window)
-        info["focused"] = self._window_to_id.get(window, -1)==self._has_focus
+        info["focused"] = self._has_focus and self._window_to_id.get(window, -1)==self._has_focus
+        info["grabbed"] = self._has_grab and self._window_to_id.get(window, -1)==self._has_grab
         return info
 
 
     def set_workarea(self, workarea):
         self._wm.set_workarea(workarea.x, workarea.y, workarea.width, workarea.height)
 
+    def set_desktop_geometry(self, width, height):
+        self._wm.set_desktop_geometry(width, height)
+
+    def set_dpi(self, xdpi, ydpi):
+        self._wm.set_dpi(xdpi, ydpi)
+
+
     def get_transient_for(self, window):
         return self._desktop_manager.get_transient_for(window, self._window_to_id)
 
@@ -276,6 +286,14 @@ class XpraServer(gobject.GObject, X11ServerBase):
         X11ServerBase.cleanup(self)
         cleanup_x11_filter()
         cleanup_all_event_receivers()
+        if self._wm:
+            self._wm.cleanup()
+            self._wm = None
+        if self._has_grab:
+            #normally we set this value when we receive the NotifyUngrab
+            #but at this point in the cleanup, we probably won't, so force set it:
+            self._has_grab = 0
+            self.X11_ungrab()
 
 
     def cleanup_source(self, protocol):
@@ -285,6 +303,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if had_client and not has_client:
             #last client is gone:
             self.reset_settings()
+            if self._has_grab:
+                self.X11_ungrab()
 
 
     def load_existing_windows(self, system_tray):
@@ -367,11 +387,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
         window.managed_connect("client-contents-changed", self._contents_changed)
         window.managed_connect("unmanaged", self._lost_window)
         window.managed_connect("raised", self._raised_window)
-        window.managed_connect("pointer-grab", self._pointer_grab)
-        window.managed_connect("pointer-ungrab", self._pointer_ungrab)
+        window.managed_connect("grab", self._window_grab)
+        window.managed_connect("ungrab", self._window_ungrab)
         return wid
 
-    _window_export_properties = ("title", "size-hints", "fullscreen", "maximized")
+    _window_export_properties = ("title", "size-hints", "fullscreen", "maximized", "opacity")
     def _add_new_window(self, window):
         self._add_new_window_common(window)
         for prop in self._window_export_properties:
@@ -436,6 +456,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 self._send_new_or_window_packet(window)
         except Unmanageable, e:
             if window:
+                windowlog("window %s is not manageable: %s", window, e)
                 #if window is set, we failed after instantiating it,
                 #so we need to fail it manually:
                 window.setup_failed(e)
@@ -516,11 +537,16 @@ class XpraServer(gobject.GObject, X11ServerBase):
             return
         had_focus = self._id_to_window.get(self._has_focus)
         def reset_focus():
-            focuslog("reset_focus() %s / %s had focus", self._has_focus, had_focus)
+            toplevel = None
+            if self._wm:
+                toplevel = self._wm.get_property("toplevel")
+            focuslog("reset_focus() %s / %s had focus (toplevel=%s)", self._has_focus, had_focus, toplevel)
             self._clear_keys_pressed()
             # FIXME: kind of a hack:
             self._has_focus = 0
-            self._wm.get_property("toplevel").reset_x_focus()
+            #toplevel may be None during cleanup!
+            if toplevel:
+                toplevel.reset_x_focus()
 
         if wid == 0:
             #wid==0 means root window
@@ -587,17 +613,23 @@ class XpraServer(gobject.GObject, X11ServerBase):
             self._damage(window, event.x, event.y, event.width, event.height)
 
 
-    def _pointer_grab(self, window, event):
-        log("pointer_grab(%s, %s)", window, event)
-        wid = self._window_to_id[window]
+    def _window_grab(self, window, event):
+        grab_id = self._window_to_id.get(window, -1)
+        grablog("window_grab(%s, %s) has_grab=%s, has focus=%s, grab window=%s", window, event, self._has_grab, self._has_focus, grab_id)
+        if grab_id<0 or self._has_grab==grab_id:
+            return
+        self._has_grab = grab_id
         for ss in self._server_sources.values():
-            ss.pointer_grab(wid)
+            ss.pointer_grab(self._has_grab)
 
-    def _pointer_ungrab(self, window, event):
-        log("pointer_ungrab(%s, %s)", window, event)
-        wid = self._window_to_id[window]
+    def _window_ungrab(self, window, event):
+        grab_id = self._window_to_id.get(window, -1)
+        grablog("window_ungrab(%s, %s) has_grab=%s, has focus=%s, grab window=%s", window, event, self._has_grab, self._has_focus, grab_id)
+        if not self._has_grab:
+            return
+        self._has_grab = 0
         for ss in self._server_sources.values():
-            ss.pointer_ungrab(wid)
+            ss.pointer_ungrab(grab_id)
 
 
     def _raised_window(self, window, event):
@@ -666,7 +698,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
     def _move_pointer(self, wid, pos):
         window = self._id_to_window.get(wid)
         if not window:
-            mouselog("_process_mouse_common() invalid window id: %s", wid)
+            mouselog("_move_pointer(%s, %s) invalid window id", wid, pos)
         else:
             mouselog("raising %s", window)
             window.raise_window()
@@ -814,7 +846,6 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 value = ''
                 for vk, vv in values.items():
                     value += "%s:\t%s\n" % (vk, vv)
-                value += '\n'
                 #record the actual value used
                 self._settings["resource-manager"] = value
                 v = value.encode("utf-8")
